/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  main.c
  
  ***************************************/


#include <sys/wait.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>

#include "ftp-proxy.h"
#include "common.h"

void run_daemon(void);
void run_with_inetd(void);
int accept_connection(struct sockaddr_in client);

void daemonify(void);
void write_pidfile(void);
int droppriv(void);
int init_log(void);
int do_chroot(void);

static RETSIGTYPE signal_handle(int signo);
static int listen_sockfd, noforks = 0;
static int reread_flag = FALSE;

/*Limit connections from single IP stuff*/
static struct client_info {
	u_int32_t ip;
	pid_t pid;
} *clients=NULL;
void add_client(pid_t, struct sockaddr_in client);
void rm_client(pid_t pid);

int main(int argc, char *argv[])
{
	struct sockaddr_in listen_address;

	info=NULL;
	process_cmdline(argc, argv);

	if (read_config() != 0) {
		fprintf(stderr, "Error reading configuration file\n");
		exit(1);
	}

	if (!config.inetd && config.maxforks != 0)
		daemonify();
	
	if(config.maxforksph)
		clients=malloc(sizeof(struct client_info)*config.maxforks);

	init_log();
		
	if (config.transdata) {
		sufd = transdata_setup();
		if (sufd == -1)
			config.transdata = FALSE;
	}

	if(!config.inetd) {
		write_pidfile();
		
		listen_address = config.listen_address;
		if ((listen_sockfd = listen_on_socket(&listen_address, CTRL)) == -1)
			exit(1);
		
		write_log(IMPORT, "Listening on %s:%d",
			  sstr_buf(addr2name(listen_address.sin_addr)),
			  ntohs(listen_address.sin_port));
		
		bindtodevice(listen_sockfd);
	}

	do_chroot();
	droppriv();

	signal(SIGCHLD, signal_handle);
	signal(SIGINT, signal_handle);
	signal(SIGTERM, signal_handle);
	signal(SIGHUP, signal_handle);
	signal(SIGPIPE, SIG_IGN);
	sstr_setopts(sstrerr, 0);

	if (cache_geninit() == -1) {
		fprintf(stderr, "Unable to initialise cache. Exiting\n");
		exit(1);
	}

	if(config.inetd) run_with_inetd();
	else run_daemon();

	return (0);
}

void run_with_inetd(void)
{
	struct sockaddr_in client_address;
	int len = sizeof(client_address);

	if( getpeername( 0, (struct sockaddr *)&client_address, &len ) )
	{
		debug_perr( "getpeername" );
		quit( SOCKET );
	}

	signal( SIGHUP, SIG_IGN );
	init_session( 0, client_address );
}

void run_daemon(void)
{
	struct sockaddr_in client_address;
	int len = sizeof(client_address);
	int fd;

	do {
		fd = accept(listen_sockfd,
			    (struct sockaddr *) &client_address, &len);
		if (fd < 0) {
			if(errno==EINTR) continue;
			debug_perr("accept");
			quit(SOCKET);
		}
		if (reread_flag) {
			reread_flag = FALSE;
			reread_config();
			if(config.maxforksph)
				clients = realloc(clients, sizeof(struct client_info)
						* (config.maxforksph > noforks ? 
						   config.maxforksph : noforks));
		}

		if (!accept_connection(client_address)) {
			close(fd);
		} else {
			pid_t pid;

			if (config.maxforks == 0) {
				close(listen_sockfd);
				signal(SIGHUP, SIG_IGN); 
				signal(SIGALRM, signal_handle);
				alarm(config.timeout);
				init_session(fd, client_address);
				exit(0);
			}

			pid = fork();
			switch (pid) {
			case -1:
				debug_perr("fork failed");
				close(fd);
				break;
			case 0:
				signal(SIGHUP, SIG_IGN);
				signal(SIGALRM, signal_handle);
				alarm(config.timeout);
				init_session(fd, client_address);
				exit(0);
			default:
				close(fd);
				if(config.maxforksph)
					add_client(pid, client_address);
				noforks++;
			}
		}
	} while (TRUE);

	freopen("/dev/null", "w", stdin);
	freopen("/dev/null", "w", stdout);
}

int accept_connection(struct sockaddr_in client)
{

	if (noforks >= config.maxforks && config.maxforks != 0) {
		write_log(ERROR,
			  "Connect from %s refused: Too many connections",
			  inet_ntoa(client.sin_addr));
		return (FALSE);
	}

	if (config.bdefend && ntohs(client.sin_port) == 20) {
		write_log(ATTACK,
			  "Connect from %s refused: Comes from ftp-data port",
			  inet_ntoa(client.sin_addr));
		return (FALSE);
	}

	if(config.maxforksph) {
		int i,x;
		for(x=0,i=0;x<noforks && i<config.maxforksph;x++)
			if(clients[x].ip==client.sin_addr.s_addr) i++;
		if(i>=config.maxforksph) {
			write_log(ERROR, "Connect from %s refused: "
				  "too many connections from that host",
				  inet_ntoa(client.sin_addr));
			return(FALSE);
		}
	}
	return (TRUE);
}

int droppriv(void)
{
	if (config.gid != 0) {
		setgid(config.gid);
		setgid(config.gid);
	}
	if (config.uid != 0) {
		if (config.transdata && sufd == 0) {	/*Kernel 2.2 we need root later */
			seteuid(config.uid);
			write_log(IMPORT, "Temporarily dropped privileges");
		} else {
			setuid(config.uid);
			setuid(config.uid);
			write_log(IMPORT, "Dropped privileges");
		}
	}
	return (0);
}


int init_log(void)
{
	if (config.logfile) {
		int tmpfd;
		tmpfd =
		    open(config.logfile, O_APPEND | O_CREAT | O_WRONLY,
			 S_IRUSR | S_IWUSR);
		if (tmpfd == -1 || dup2(tmpfd, 2) == -1) {
			fprintf(stderr, "Unable to open logfile %s\n",
				config.logfile);
			return (-1);
		}
	}
	openlog("frox", LOG_PID | LOG_CONS | LOG_NDELAY, LOG_DAEMON);
	return (0);
}

int do_chroot(void)
{
	if (!config.chroot)
		return (0);
	if (chroot(config.chroot) != 0 || chdir("/") != 0) {
		free(config.chroot);
		config.chroot = NULL;
		return (-1);
	}
	write_log(IMPORT, "Chrooted to %s", config.chroot);

	strip_filenames();

	return (0);
}

void daemonify()
{
	switch (fork()) {
	case -1:
		write_log(ERROR, "can't fork daemon");
		exit(-1);
		break;
	case 0:
		/******** child ********/
		break;
	default:
		/******** parent ********/
		exit(0);
	}

	debug("Forked to background\n");
}

void write_pidfile(void)
{
	FILE *fp;
	pid_t pid;

	pid = getpid();

	if (config.pidfile) {
		fp = fopen(config.pidfile, "w");
		if (fp != NULL) {
			fprintf(fp, "%d\n", pid);
			fclose(fp);
		}
	}
}

void add_client(pid_t pid, struct sockaddr_in client)
{
	int i;

	clients[noforks].pid=pid;
	clients[noforks].ip=client.sin_addr.s_addr;

	for(i=0;i<=noforks;i++)
		printf("pid: %d ip: %x\n", clients[i].pid, clients[i].ip);

}

void rm_client(pid_t pid)
{
	int i;

	for(i=0;i<noforks && pid!=clients[i].pid; i++);
	if(i<noforks) {
		clients[i].pid=clients[noforks].pid;
		clients[i].ip=clients[noforks].ip;
	}
}

static RETSIGTYPE signal_handle(int signo)
{
	pid_t pid;
	switch (signo) {
	case SIGCHLD:
		if(info) return; /*Children do their own waitpid()*/
		while((pid=waitpid(-1, (int *) 0, WNOHANG))>0){
			noforks--;
			if(config.maxforksph) rm_client(pid);
		}
#ifdef TRANS_DATA
		if (noforks == 0 && config.transdata) {
			mangle_req req;
			req.type = FLUSH;
			write(sufd, &req, sizeof(req));
		}
#endif
		
		signal(signo, signal_handle);
		break;
	case SIGINT:
	case SIGTERM:
		close(listen_sockfd);
		exit(0);
	case SIGHUP:		/*Reread config file. Not safe from signal handler. */
		reread_flag = TRUE;
		signal(signo, signal_handle);
		break;
	case SIGALRM:
		quit(TIMEOUT);
	}
}
