/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    misc.c - Miscellaneous stuff!
  ***************************************/


#include <stdarg.h>
#include <syslog.h>
#include <netdb.h>
#include <stdlib.h>

#include "common.h"
#include "control.h"

/* ------------------------------------------------------------- **
** Listens on socket. If port==0 then picks a port from the range
** specified in the config file for the use given in use.
** ------------------------------------------------------------- */
int listen_on_socket(struct sockaddr_in *listen_address, socketuse use)
{
	int sockfd;

	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (use == CTRL) {
		int i = 1;
		setsockopt(sockfd, SOL_SOCKET, SO_OOBINLINE, &i, sizeof(i));
		i = i;
		setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i));
	}

	if (listen_address->sin_port) {
		if (bind(sockfd, (struct sockaddr *) listen_address,
			 sizeof(*listen_address))) {
			debug_perr("bind");
			close(sockfd);
			return (-1);
		}
	} else {
		int lo, hi;

		lo = config_loport(use);
		hi = config_hiport(use);

		if (bind_me(sockfd, listen_address, lo, hi, hi - lo)) {
			debug_perr("bind_me");
			close(sockfd);
			return (-1);
		}
	}
	if (listen(sockfd, 5)) {
		close(sockfd);
		return (-1);
	}

	return (sockfd);
}

/* ------------------------------------------------------------- **
** Connects to address. Local port is picked from the range given in
** the config file for the use specified in use. 
** ------------------------------------------------------------- */
int connect_to_socket(struct sockaddr_in connect_address, socketuse use)
{
	int sockfd;
	struct sockaddr_in address;
	int lo, hi;

	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		debug_perr("socket");
		quit(SOCKET);
	}

	lo = config_loport(use);
	hi = config_hiport(use);

	memset(&address, 0, sizeof(address));
	address.sin_family = AF_INET;

	if (bind_me(sockfd, &address, lo, hi, hi - lo)) {
		debug_perr("bind_me");
		close(sockfd);
		return (-1);
	}

	connect_address.sin_family = AF_INET;
	if (connect(sockfd, (struct sockaddr *) &connect_address,
		    sizeof(address))) {
		debug_perr("connect");
		close(sockfd);
		return (-1);
	}

	return (sockfd);
}


/* ------------------------------------------------------------- **
** Try "attempts" times to bind socket "fd" to address "address", with
** a port picked at random from between "lo" and "hi".
** ------------------------------------------------------------- */
int bind_me(int fd, struct sockaddr_in *address, int lo, int hi, int attempts)
{
	int i, j;

	srand(time(NULL));

	for (i = 0; i <= attempts; i++) {
		j = (rand() % (hi - lo)) + lo;
		address->sin_port = htons(j);

		if (bind(fd, (struct sockaddr *) address,
			 sizeof(*address)) == 0)
			break;
		if (errno != EADDRINUSE) {
			return (-1);
		}
	}

	return (i > attempts ? -1 : 0);
}

/* ------------------------------------------------------------- **
** Convert a comma separted values address:port to a sockaddr_in 
** in network order
** ------------------------------------------------------------- */
struct sockaddr_in com2n(int a1, int a2, int a3, int a4, int p1, int p2)
{
	struct sockaddr_in ret;

	ret.sin_addr.s_addr = htonl(a4 + (a3 << 8) + (a2 << 16) + (a1 << 24));
	ret.sin_port = htons((p1 << 8) + p2);
	ret.sin_family = AF_INET;

	return (ret);
}


/* ------------------------------------------------------------- **
** Convert network order address to comma separated values
** ------------------------------------------------------------- */
void n2com(struct sockaddr_in address, int *a1, int *a2, int *a3, int *a4,
	   int *p1, int *p2)
{
	address.sin_addr.s_addr = ntohl(address.sin_addr.s_addr);
	address.sin_port = ntohs(address.sin_port);
	*a1 = (address.sin_addr.s_addr & 0xFF000000) >> 24;
	*a2 = (address.sin_addr.s_addr & 0x00FF0000) >> 16;
	*a3 = (address.sin_addr.s_addr & 0x0000FF00) >> 8;
	*a4 = (address.sin_addr.s_addr & 0x000000FF);
	*p1 = (address.sin_port & 0xFF00) >> 8;
	*p2 = (address.sin_port & 0x00FF);
}

/* ------------------------------------------------------------- **
** Extract a comma delimited address/port from buf. buf is NULL
** terminated, but otherwise unchecked user input --- make sure we
** don't buffer overflow here!
** ------------------------------------------------------------- */
struct sockaddr_in extract_address(sstr * p)
{
	int i, a[6];
	struct sockaddr_in tmp;

	tmp.sin_addr.s_addr = tmp.sin_port = 0;

	sstr_split(p, NULL, 0, sstr_pbrk2(p, "01234456789"));
	for (i = 0; i < 6; i++) {
		a[i] = sstr_atoi(p);
		if (i != 5 && sstr_token(p, NULL, ",", 0) == -1)
			return (tmp);
	}

	tmp = com2n(a[0], a[1], a[2], a[3], a[4], a[5]);
	return (tmp);
}


void write_log(int priority, char *msg, ...)
{
	char buf[MAX_LINE_LEN];
	sstr *s;
	va_list argptr;
	time_t tstamp;

	if (priority > config.loglevel)
		return;
	va_start(argptr, msg);
	vsnprintf(buf, MAX_LINE_LEN - 10, msg, argptr);
	va_end(argptr);
	buf[MAX_LINE_LEN - 9] = 0;

	if (priority <= ERROR) {
		syslog(LOG_ERR | LOG_DAEMON, "%s\n", buf);
	} else {
		syslog(LOG_NOTICE | LOG_DAEMON, "%s\n", buf);
	}

	s = sstr_init(MAX_LINE_LEN);
	time(&tstamp);
	sstr_cpy2(s, ctime(&tstamp));
	sstr_setchar(s, sstr_chr(s, '\n'), ' ');
	sstr_apprintf(s, "frox[%d] %s\n", getpid(), buf);
	sstr_write(2, s, 0);
	sstr_free(s);
}

/* ------------------------------------------------------------- **
** Return hostname of address, or failing that the IP as a string
** ------------------------------------------------------------- */
sstr *addr2name(struct in_addr address)
{
	struct hostent *hostinfo;
	static sstr *buf = NULL;

	if (!buf)
		buf = sstr_init(MAX_LINE_LEN);
	sstr_cpy2(buf, inet_ntoa(address));

	if ((hostinfo = gethostbyaddr((char *) &address,
				      sizeof(address), AF_INET)) != NULL)
		sstr_apprintf(buf, "(%s)", hostinfo->h_name);

	return (buf);
}

/* ------------------------------------------------------------- **
** Deal with session quit conditions.
** ------------------------------------------------------------- */
void quit(exitreason reason)
{
	il_free();

	switch (reason) {
	case TIMEOUT:
		write_log(INFO, "Select timed out.");
		closeandexit(0);
	case SERVER_CLOSE:
		write_log(INFO, "Server closed the control connection.");
		closeandexit(0);
	case CLIENT_CLOSE:
		write_log(INFO, "Client closed the control connection.");
		closeandexit(0);

	case NO_SERVER_DATA:
		write_log(ERROR, "  Unable to connect to server data port");
		closeandexit(-1);
	case NO_CLIENT_DATA:
		write_log(ERROR, "  Unable to connect to client data port");
		closeandexit(-1);
	case MALLOC:
		write_log(ERROR, "  malloc failed");
		closeandexit(-1);
	case SOCKET:
		write_log(ERROR, "  socket command failed");
		closeandexit(-1);
	case CLIENT_RUBBISH:
		send_cmessage(421, "You are sending me rubbish. Goodbye.");
		write_log(ATTACK,
			  "Client is sending us a badly formed control stream.");
		closeandexit(-1);
	case SERVER_RUBBISH:
		send_cmessage(421,
			      "FTP server is sending you rubbish! Closing connection.");
		write_log(ATTACK,
			  "Server is sending us a badly formed control stream.");
		closeandexit(-1);
	}
}


void closeandexit(int status)
{
	close(info->client_control.fd);
	close(info->client_data.fd);
	close(info->server_control.fd);
	close(info->server_data.fd);
	close(info->listen);
	free(info);
	write_log(INFO, "Closing session");
	exit(status);
}

void sstrerr(void)
{
	write_log(ERROR, "sstr internal failure. Exiting");
	quit(-1);
}
